// Copyright (c) 2017, Tobias Mueller tm(at)tm3d.de
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met:
//
//  * Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//  * Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the
//    distribution.
//  * All advertising materials mentioning features or use of this
//    software must display the following acknowledgement: This product
//    includes software developed by tm3d.de and its contributors.
//  * Neither the name of tm3d.de nor the names of its contributors may
//    be used to endorse or promote products derived from this software
//    without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#ifndef __OWINTERFACES_H_
#define __OWINTERFACES_H_

#if defined(WIN) || defined(LINUX)
#else
#if defined(_WINDOWS) || defined(__WINDOWS__) || defined(_WIN32) || defined(WIN32)
#define WIN
#else
#define LINUX
#endif
#endif


#include <vector>
#include <stdint.h>
#include <string>
#ifdef LINUX
#include <termios.h>
#endif
#include <stdarg.h>

#ifdef WIN
#include <Windows.h>

#endif

class owDevice;
class owDeviceConfig;
#include "owDeviceConfig.h"

#if defined(_WINDOWS) || defined(__WINDOWS__) || defined(_WIN32) || defined(WIN32)
   #define USE_WINDOWS_TIME 1
   #include <windows.h>
   //#include "win32/usb.h" // libusb header
#else
   #define USE_WINDOWS_TIME 0
   #include <time.h>
   #include <sys/time.h>

   #include <usb.h>
#endif

#ifndef FALSE
#define FALSE 0
#endif
#ifndef TRUE
#define TRUE 1
#endif

#define OWLOG_INFO 1
#define OWLOG_WARNING 2
#define OWLOG_ERROR 3

#ifdef LINUX

#define BLUE_B "\033[1;34m"
#define COLOR_E "\033[0m"
#define C2_B "\033[0;36m"
#define C3_B "\033[3;34m"
#define C4_B "\033[1;33m"
#define C5_B "\033[4;33m"
#define CLE_B "\033[1;31m"
#define CLW_B "\033[1;33m"
#endif
#ifdef WIN
#define BLUE_B ""
#define C2_B ""
#define C3_B ""
#define C4_B ""
#define COLOR_E ""
#define C5_B ""
#define CLE_B ""
#define CLW_B ""

#endif

class owLog {
	int lcount;
	std::string logtext;
	int level;
	int loglevel;
		
public:
	owLog() {
		lcount=0;
		loglevel=0;
	} 
	virtual int set(int llevel,const char *format, ...) {
		char s[300];
		va_list arg;
		int done;
		va_start (arg, format);
#ifdef LINUX
		done = vsprintf (s, format, arg);
#endif
#ifdef WIN
		done = vsprintf_s(s,300, format, arg);
#endif
		va_end (arg);
		logtext=s;
		level=llevel;
		lcount++;
		if (level>=loglevel) {
			if (loglevel==OWLOG_ERROR) printf(CLE_B "%s" COLOR_E "\n",s);
			else if (loglevel==OWLOG_WARNING) printf(CLW_B "%s" COLOR_E "\n",s);
			else printf("%s\n",s);
		}
		return done;
	}
	virtual void clear() {
		lcount=0;
	}
	virtual int last() {
		if (lcount) return level; else return 0;
	}
	virtual void setLogLevel(int llevel ){
		loglevel=llevel;
	}
	
};


typedef union  {
	uint8_t byte[8];
	uint64_t num;
} snum_t;

class owInterface {
protected:
	int block;
	uint8_t crc8;
	uint16_t crc16;
	uint8_t docrc8(uint8_t value);
	uint16_t docrc16(uint16_t cdata);
	
	// global search state
	unsigned char ROM_NO[8];
	int LastDiscrepancy;
	int LastFamilyDiscrepancy;
	int LastDeviceFlag;
	
	int devices_changed;
	
public:
	std::vector<snum_t> device_nums;
	std::vector<owDevice*> devices;
	int maxrepeat;
	owLog *log;
	owInterface() {
		block=0;
		log=new owLog();
		maxrepeat=4;
		devices_changed=0;
	}
	void Wait_Free() {
		while (block) {};
		block = 1;
	}
	void Free() {
		block = 0;
	}	
	
	int isDevicesChanged() {
		int dc=devices_changed;
		devices_changed=0;
		return dc;
	}
	virtual void Clean();
	virtual int Find() ;
	virtual int InitAdapter(std::string s) {
		uint8_t param=atoi(s.c_str());
		return InitAdapter(param);
	}
	virtual void ReleaseAdapter() {}
	virtual int InitAdapter(uint8_t)=0;
	//Return 1 if Devices Present
	virtual int Reset()=0;
	virtual uint8_t sendrecivBit(uint8_t bit) =0;
	


	virtual uint8_t sendrecivByte(uint8_t byte) {
		uint8_t b=byte;
		uint8_t rb=0;
		for(int i=0;i<8;i++) {
       		rb=rb>>1;
			rb|=sendrecivBit(b&1)<<7;
                b=b>>1;
		}
		return rb;
	}
	
	virtual int MatchRom(snum_t snum);
	virtual int Communicate(std::vector<uint8_t> *data, int scount, int rcount);
	
	

	uint8_t calcCRC8(std::vector<uint8_t> data);
	uint16_t calcCRC16(std::vector<uint8_t> data);
	int testCRC16(std::vector<uint8_t> data);
		
	void resetFlasher();
	void resetID();
	int programmPage(int pagenr, std::vector<uint8_t> page, int pagesize);
	int flashHEXFile(std::string filename,snum_t dev,int resetid,int progress);
	
	
	virtual int owFirst() {
		LastDiscrepancy = 0;
		LastDeviceFlag = FALSE;
		LastFamilyDiscrepancy = 0;
		return owSearch();		
	}
	virtual int owNext()  {
		return owSearch();
	}
	virtual ~owInterface()  {
		ReleaseAdapter();
	}
protected:
	int owSearch();
	
};

#define OWDCS_NONE 0
#define OWDCS_DEFAULT 1
#define OWDCS_16 2
#define OWDCS_24 3


class owDevice {
protected:
	owInterface* owi;
	snum_t snum;
	std::vector<int32_t> raw;
public:
	owDeviceConfig *config;
	int configstate;
	std::vector<double> values;
	int lastfound; //0 found in last searchrom, 1 found in the searchrom befor, and so on
	owDevice(owInterface *owi_,snum_t num);
	//return 1 for real Config, return 0 for DefaultConig
	int readConfig();
	virtual void setDefaultConfig();
		//Information zum owDevice	
	std::vector<std::string> getFamilyInfo();
	
	snum_t getNum() {return snum;}

	int Communicate(std::vector<uint8_t> *data, int scount, int rcount);
	int CommunicateShort(std::vector<uint8_t> *data, int scount, int rcount);
	void changeID(snum_t nid);
	void runFlasher();
	virtual int convertAll() {return 1;}
	virtual int readMemory(int page,int start, int count,std::vector<uint8_t> *data) {return 0;};
	virtual int writeMemory(int page,int start, int count,std::vector<uint8_t> *data) {return 0;};
	virtual ~owDevice() {}
};

class owDeviceDS18B20:public owDevice {
public:
	owDeviceDS18B20(owInterface *owi_,snum_t num):owDevice(owi_,num) {	}
	virtual void setDefaultConfig();
	int readScratchpad(std::vector<uint8_t> *sp);
	virtual int convertAll();
} ;


class owDeviceDS2438:public owDevice {
public:
	owDeviceDS2438(owInterface *owi_,snum_t num):owDevice(owi_,num) {}
	virtual void setDefaultConfig();

	int setConfigByte(uint8_t cb);
	int readScratchpad(std::vector<uint8_t> *sp, uint8_t page, int recall);
	virtual int readMemory(int page,int start, int count,std::vector<uint8_t> *data);
	virtual int writeMemory(int page,int start, int count,std::vector<uint8_t> *data);
	virtual int convertAll();

} ;
class owDeviceDS2450:public owDevice {
public:
	owDeviceDS2450(owInterface *owi_,snum_t num):owDevice(owi_,num) {}

	virtual void setDefaultConfig();
	virtual int convertAll();
	void readMemory(uint8_t addr,std::vector<uint8_t> *sp);
	void writeMemory(uint8_t addr,std::vector<uint8_t> *sp);
	void convert(uint8_t mask, uint8_t preset); 

} ;


class owDeviceDS2423:public owDevice {
public:
	owDeviceDS2423(owInterface *owi_,snum_t num):owDevice(owi_,num) {}

	virtual void setDefaultConfig();
	virtual int convertAll();
	uint32_t readCounter(uint8_t page);

} ;


#endif
